#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)

from __future__ import annotations

import threading
import traceback
from typing import TYPE_CHECKING

import os
from enum import Enum

from fim.graph.neo4j_property_graph import Neo4jGraphImporter

from fabric_cf.actor.core.common.exceptions import ContainerException
from fabric_cf.actor.core.container.remote_actor_cache import RemoteActorCacheSingleton
from fabric_cf.actor.core.container.db.container_database import ContainerDatabase
from fabric_cf.actor.core.proxies.actor_location import ActorLocation
from fabric_cf.actor.core.proxies.proxy_factory import ProxyFactorySingleton
from fabric_cf.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.util.reflection_utils import ReflectionUtils
from fabric_cf.actor.core.apis.abc_actor_container import ABCActorContainer
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.container.protocol_descriptor import ProtocolDescriptor
from fabric_cf.actor.core.kernel.kernel_tick import KernelTick
from fabric_cf.actor.core.manage.management_object_manager import ManagementObjectManager
from fabric_cf.actor.core.util.id import ID


if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.apis.abc_tick import ABCTick
    from fabric_cf.actor.core.apis.abc_actor_identity import ABCActorIdentity
    from fabric_cf.actor.core.apis.abc_container_database import ABCContainerDatabase
    from fabric_cf.actor.boot.configuration import Configuration


class ContainerState(Enum):
    """
    Container State enumeration
    """
    Unknown = 1
    Starting = 2
    Recovering = 3
    Started = 4
    Stopping = 5
    Stopped = 6
    Failed = 7


class Container(ABCActorContainer):
    """
    Container is the "heart" of Fabric Core system. The container manager is responsible for managing the core instance.
    This is a singleton class.
    """

    PropertyBeginningOfTime = "BeginningOfTime"
    PropertyCycleMillis = "CycleMillis"
    PropertyManualTicks = "ManualTicks"

    PropertyTime = "time"

    def __init__(self):
        self.state = ContainerState.Unknown
        self.guid = None
        self.config = None
        self.ticker = None
        self.protocols = {}
        self.fresh = True
        self.recovered = False
        self.management_object_manager = ManagementObjectManager()
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.db = None
        self.container_lock = threading.Lock()
        self.actor = None

    def get_actor(self) -> ABCActorMixin:
        """
        Return the actor running in the container
        @return actor
        """
        return self.actor

    def cleanup_neo4j(self):
        """
        Cleanup Neo4j on clean restart
        """
        self.logger.debug("Cleanup Neo4j database started")
        config = self.config.get_global_config().get_neo4j_config()
        neo4j_graph_importer = Neo4jGraphImporter(url=config["url"], user=config["user"],
                                                  pswd=config["pass"],
                                                  import_host_dir=config["import_host_dir"],
                                                  import_dir=config["import_dir"])
        neo4j_graph_importer.delete_all_graphs()
        self.logger.debug("Cleanup Neo4j database completed")

    def determine_boot_mode(self):
        """
        Determine boot mode is clean restart or stateful restart
        """
        filename = Constants.SUPERBLOCK_LOCATION
        self.logger.debug(f"Checking if this container is recovering. Looking for: {filename}")
        if os.path.isfile(filename):
            self.logger.debug("Found super block file. This container is recovering")
            self.fresh = False
        else:
            self.logger.debug("Super block file does not exist. This is a fresh container")
            self.fresh = True

    def create_super_block(self):
        """
        Create Super Block
        """
        self.logger.debug("Creating superblock")
        file = None
        try:
            file = open(Constants.SUPERBLOCK_LOCATION, 'r')
        except IOError:
            file = open(Constants.SUPERBLOCK_LOCATION, 'w')
        finally:
            if file is not None:
                file.close()

    def boot(self):
        """
        Startup a container
        """
        self.logger.debug("Booting")
        self.boot_common()
        if self.is_fresh():
            self.logger.debug("Booting a fresh container")
            self.boot_basic()
            self.finish_fresh_boot()
        else:
            self.logger.debug("Recovering an existing container")
            self.recover_basic()
            self.recover_actors()
            self.finish_recovery_boot()
        self.write_state_file()

    def create_database(self):
        """
        Create Database
        """
        user = self.config.get_global_config().get_database().get(Constants.PROPERTY_CONF_DB_USER, None)
        password = self.config.get_global_config().get_database().get(Constants.PROPERTY_CONF_DB_PASSWORD, None)
        dbname = self.config.get_global_config().get_database().get(Constants.PROPERTY_CONF_DB_NAME, None)
        dbhost = self.config.get_global_config().get_database().get(Constants.PROPERTY_CONF_DB_HOST, None)
        self.db = ContainerDatabase(user=user, password=password, database=dbname, db_host=dbhost, logger=self.logger)
        if self.is_fresh():
            self.db.set_reset_state(value=True)
        else:
            self.db.set_reset_state(value=False)
        self.db.initialize()

    def initialize(self, *, config: Configuration):
        """
        Initialize container and actor
        """
        if config is None:
            raise ContainerException("handlers cannot be null")

        with self.container_lock:
            if self.state != ContainerState.Unknown:
                raise ContainerException(f"Cannot initialize container in state: {self.state}")

            self.state = ContainerState.Starting

        self.config = config
        failed = False

        try:
            self.determine_boot_mode()
            self.create_database()
            self.boot()

            if self.is_fresh():
                try:
                    self.cleanup_neo4j()
                    from fabric_cf.actor.boot.configuration_loader import ConfigurationLoader
                    loader = ConfigurationLoader()
                    loader.process(config=self.config)
                except Exception as e:
                    self.logger.error(traceback.format_exc())
                    self.logger.error(f"Failed to instantiate actors {e}")
                    self.logger.error("This container may need to be restored to a clean state")
                    raise e
                # Create State file only after successful fresh boot
                self.create_super_block()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(e)
            failed = True
            raise e
        finally:
            with self.container_lock:
                if failed:
                    self.state = ContainerState.Failed
                else:
                    self.state = ContainerState.Started

    def boot_common(self):
        """
        Perform common boot actions
        """
        self.logger.debug("Performing common boot tasks")
        self.define_protocols()
        self.management_object_manager.initialize(db=self.db)

    def define_protocols(self):
        """
        Define Protocols i.e. Kafka and Local
        """
        self.logger.debug("Defining container protocols")
        desc = ProtocolDescriptor(protocol=Constants.PROTOCOL_LOCAL)
        self.logger.debug(f"Registering protocol {Constants.PROTOCOL_LOCAL}")
        self.register_protocol(protocol=desc)

        kafka_topic_name = self.get_config().get_actor().get_kafka_topic()
        self.logger.debug("Kafka Topic {}".format(kafka_topic_name))
        if kafka_topic_name is not None:
            desc = ProtocolDescriptor(protocol=Constants.PROTOCOL_KAFKA, location=kafka_topic_name)
            self.logger.debug(f"Registering protocol {Constants.PROTOCOL_KAFKA}")
            self.register_protocol(protocol=desc)

    def boot_basic(self):
        """
        Perform basic boot actions
        """
        self.guid = self.get_config().get_global_config().get_container().get(Constants.PROPERTY_CONF_CONTAINER_GUID,
                                                                              None)
        self.logger.debug(f"Container guid is {self.guid}")
        self.set_time()
        self.persist_basic()

    def persist_basic(self):
        """
        Save basic and time information
        """
        self.persist_container()
        self.persist_time()

    def persist_container(self):
        """
        Save container information in database
        """
        properties = {Constants.PROPERTY_CONF_CONTAINER_GUID: self.guid}
        self.db.add_container_properties(properties=properties)

    def persist_time(self):
        """
        Save time in database
        """
        properties = {self.PropertyTime: self.PropertyTime,
                      self.PropertyBeginningOfTime: self.ticker.get_beginning_of_time(),
                      self.PropertyCycleMillis: self.ticker.get_cycle_millis(),
                      self.PropertyManualTicks: self.ticker.is_manual()}
        self.db.add_time(properties=properties)

    def set_time(self):
        """
        Set the Actor clock and time
        """
        start_time = int(self.config.get_global_config().get_time().get(Constants.PROPERTY_CONF_TIME_START_TIME, None))
        if start_time == -1:
            start_time = ActorClock.get_current_milliseconds()

        cycle_millis = int(self.config.get_global_config().get_time().get(Constants.PROPERTY_CONF_TIME_CYCLE_MILLIS,
                                                                          None))

        manual = False
        if self.config.get_global_config().get_time().get(Constants.PROPERTY_CONF_TIME_MANUAL, None):
            manual = True

        self.create_and_start_tick(start_time=start_time, cycle_millis=cycle_millis, manual=manual)

    def create_and_start_tick(self, *, start_time: int, cycle_millis: int, manual: bool):
        """
        Create Actor clock and start the ticker
        """
        self.logger.debug("Creating container ticker")
        self.ticker = KernelTick()
        self.ticker.set_beginning_of_time(value=start_time)
        self.ticker.set_cycle_millis(cycle_millis=cycle_millis)
        self.ticker.set_manual(value=manual)
        self.ticker.initialize()
        self.ticker.start()

        if not self.ticker.is_manual():
            self.logger.info(f"Using automatic ticks. Tick length={self.ticker.get_cycle_millis()} ms")
        else:
            self.logger.info(f"Using manual ticks. Tick length={self.ticker.get_cycle_millis()} ms")

    def finish_fresh_boot(self):
        """
        Complete fresh boot
        """
        self.create_container_manager_object()

    def finish_recovery_boot(self):
        """
        Complete recovery boot
        """

    def create_container_manager_object(self):
        """
        Create Container Manager Object
        """
        self.logger.info("Creating container manager object")
        from fabric_cf.actor.core.manage.container_management_object import ContainerManagementObject
        management_object = ContainerManagementObject()
        self.management_object_manager.register_manager_object(manager=management_object)

    def recover_basic(self):
        """
        Recover Basic entities such as Container GUID and time
        """
        if self.is_fresh():
            raise ContainerException("A fresh container cannot be recovered")
        try:
            self.container_lock.acquire()
            if self.state != ContainerState.Starting:
                raise ContainerException(f"Invalid state for recovery: {self.state}")
            self.state = ContainerState.Recovering
        finally:
            self.container_lock.release()

        self.recover_guid()
        self.recover_time()

    def recover_guid(self):
        """
        Recover GUID from database
        """
        self.logger.debug("Recovering container GUID")
        properties = self.db.get_container_properties()
        stored_guid = properties.get(Constants.PROPERTY_CONF_CONTAINER_GUID, None)
        if stored_guid is None:
            raise ContainerException("Could not obtain saved container GUID from database")
        self.guid = ID(uid=stored_guid)
        self.logger.info(f"Recovered container guid: {self.guid}")

    def recover_time(self):
        """
        Recover time from database
        """
        self.logger.debug("Recovering container time settings")
        time_obj = self.db.get_time()
        if time_obj is None:
            raise ContainerException("Could not obtain container saved state from database")
        beginning_of_time = time_obj.get(self.PropertyBeginningOfTime, None)
        cycle_millis = time_obj.get(self.PropertyCycleMillis, None)
        manual = time_obj.get(self.PropertyManualTicks, None)

        self.create_and_start_tick(start_time=beginning_of_time, cycle_millis=cycle_millis, manual=manual)

    def recover_actors(self):
        """
        Recover actors from Database (there is only one actor)
        """
        self.logger.info("Recovering actors")
        actor_list = self.db.get_actors()
        for a in actor_list:
            try:
                self.recover_actor(actor=a)
            except Exception as e:
                self.logger.error(f"Could not recover actor {a.get_name()}, exception: {e}")
                raise e

    def recover_actor(self, *, actor: ABCActorMixin):
        """
        Recover actor
        """
        actor_name = actor.get_name()
        actor.set_logger(logger=self.logger)
        self.logger.info(f"Recover actor: {actor_name}")
        self.logger.debug("Initializing the actor object")
        actor.initialize()
        # By now we have a valid actor object. We need to register it with
        # the container and call recovery for its reservations.
        self.register_recovered_actor(actor=actor)
        self.logger.debug(f"Starting recovery from database for actor {actor_name}")
        actor.recover()
        self.register(tickable=actor)

        self.logger.info(f"Actor {actor_name} recovered successfully")

    def register(self, *, tickable: ABCTick):
        """
        Register Actor with timer thread
        @param tickable actor
        """
        self.ticker.add_tickable(tickable=tickable)

    def unregister(self, *, tickable: ABCTick):
        """
        Un-Register Actor with timer thread
        @param tickable actor
        """
        self.ticker.remove_tickable(tickable=tickable)

    def start_actor(self):
        self.logger.debug("Starting the Actor")
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        producer = GlobalsSingleton.get().get_kafka_producer_with_poller(actor=self.actor)
        from fabric_cf.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
        RPCManagerSingleton.get().set_producer(producer=producer)
        RPCManagerSingleton.get().start()
        self.actor.start()
        self.logger.debug(f"Actor {self.actor.get_name()} started")

    def register_actor(self, *, actor: ABCActorMixin):
        """
        Registers a new actor: adds the actor to the database, registers actor proxies and callbacks. Must not
        register the actor with the clock! Clock registration is a separate phase.
        @param actor actor
        @raises Exception in case of error
        """
        self.db.add_actor(actor=actor)
        actor.actor_added()
        self.register_management_object(actor=actor)
        self.register_common(actor=actor)
        self.actor = actor
        self.start_actor()

    def unregister_actor(self, *, actor: ABCActorMixin):
        """
        Unregisters the actor from the container.
        @param actor actor
        @raises Exception in case of error
        """
        self.management_object_manager.unload_actor_manager_objects(actor_name=actor.get_name())
        ActorRegistrySingleton.get().unregister(actor=actor)

    def register_common(self, *, actor: ABCActorMixin):
        """
        Performs the common steps required to register an actor with the container.
        @param actor actor
        @raises Exception in case of error
        """
        ActorRegistrySingleton.get().register_actor(actor=actor)
        self.register_proxies(actor=actor)

        RemoteActorCacheSingleton.get().register_with_registry(actor=actor)

    def register_management_object(self, *, actor: ABCActorMixin):
        """
        Register Actor Management Object
        """
        module_name = actor.get_management_object_module()
        class_name = actor.get_management_object_class()
        mo = ReflectionUtils.create_instance(module_name=module_name, class_name=class_name)
        mo.set_actor(actor=actor)
        mo.initialize()
        self.management_object_manager.register_manager_object(manager=mo)

    def remove_actor(self, *, actor_name: str):
        """
        Remove actor metadata
        @param actor_name actor name;
        @raises Exception in case of error
        """
        self.db.remove_actor(actor_name=actor_name)
        actor = ActorRegistrySingleton.get().get_actor(actor_name_or_guid=actor_name)
        if actor is not None:
            ActorRegistrySingleton.get().unregister(actor=actor)
        actor.actor_removed()

    def remove_actor_database(self, *, actor_name: str):
        """
        Remove actor database
        @param actor_name actor name;
        @raises Exception in case of error
        """
        self.db.remove_actor(actor_name=actor_name)

    def register_protocol(self, *, protocol: ProtocolDescriptor):
        self.protocols[protocol.get_protocol()] = protocol
        self.logger.debug(f"Registered container protocol: {protocol.get_protocol()}")

    def register_proxies(self, *, actor: ABCActorMixin):
        """
        Registers all proxies for the specified actor.
        @param actor actor
        @raises Exception in case of error
        """
        self.logger.debug(f"Registering proxies for actor: {actor.get_name()}")

        for protocol in self.protocols.values():
            self.logger.debug(f"Processing protocol {protocol.get_protocol()}")
            location = ActorLocation()
            location.set_descriptor(descriptor=protocol)

            proxy = ProxyFactorySingleton.get().new_proxy(protocol=protocol.get_protocol(),
                                                          identity=actor.get_identity(), location=location)
            if proxy is not None:
                self.logger.debug(f"Registering proxy {actor.get_name()} for protocol {protocol.get_protocol()}")
                ActorRegistrySingleton.get().register_proxy(proxy=proxy)

            callback = ProxyFactorySingleton.get().new_callback(protocol=protocol.get_protocol(),
                                                                identity=actor.get_identity(), location=location)
            if callback is not None:
                self.logger.debug(f"Registering callback {actor.get_name()} for protocol {protocol.get_protocol()}")
                ActorRegistrySingleton.get().register_callback(callback=callback)

    def register_recovered_actor(self, *, actor: ABCActorMixin):
        """
        Registers a recovered actor.
        @param actor recovered actor
        @raises Exception in case of error
        """
        self.logger.debug("Registering a recovered actor")
        actor.actor_added()
        self.register_common(actor=actor)
        self.management_object_manager.load_actor_manager_objects(actor_name=actor.get_name())
        self.actor = actor
        self.start_actor()

    def stop(self):
        if self.ticker is not None:
            self.ticker.stop()
        else:
            raise ContainerException("The container does not have a clock")

    def tick(self):
        if self.ticker is not None:
            self.ticker.tick()
        else:
            raise ContainerException("The container does not have a clock")

    def shutdown(self):
        try:
            self.logger.info("Actor container shutting down")
            self.remove_state_file()
            self.stop()
            self.logger.info("Stopping RPC Manager")
            from fabric_cf.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
            RPCManagerSingleton.get().stop()
            self.logger.info("Stopping actors")
            actors = ActorRegistrySingleton.get().get_actors()
            for actor in actors:
                self.logger.info(f"Stopping actor: {actor.get_name()}")
                actor.stop()
                self.unregister_actor(actor=actor)

            ActorRegistrySingleton.get().clear()
            self.logger.info("Container is no longer active")
        except Exception as e:
            self.logger.error(f"Exception occurred while shutting down e: {e}")
            self.logger.error(traceback.format_exc())

    @staticmethod
    def write_state_file():
        try:
            file = open(Constants.STATE_FILE_LOCATION, 'r')
        except IOError:
            file = open(Constants.STATE_FILE_LOCATION, 'w')
        finally:
            file.close()

    @staticmethod
    def remove_state_file():
        os.remove(Constants.STATE_FILE_LOCATION)

    def get_actor_clock(self) -> ActorClock:
        if self.ticker is None:
            raise ContainerException("No tick")
        return ActorClock(beginning_of_time=self.ticker.get_beginning_of_time(),
                          cycle_millis=self.ticker.get_cycle_millis())

    def get_current_cycle(self):
        if self.ticker is None:
            return -1
        return self.ticker.get_current_cycle()

    def is_manual_clock(self) -> bool:
        result = False
        if self.ticker is not None:
            result = self.ticker.is_manual()
        return result

    def get_guid(self) -> ID:
        return self.guid

    def get_database(self) -> ABCContainerDatabase:
        return self.db

    def get_config(self) -> Configuration:
        return self.config

    def is_fresh(self) -> bool:
        return self.fresh

    def is_recovered(self):
        return self.recovered

    def get_management_object_manager(self) -> ManagementObjectManager:
        return self.management_object_manager

    def get_management_object(self, *, key: ID):
        if key is None:
            raise ContainerException("key cannot be null")
        return self.management_object_manager.get_management_object(key=key)

    def get_protocol_descriptor(self, *, protocol: str) -> ProtocolDescriptor:
        return self.protocols.get(protocol, None)

    @staticmethod
    def get_proxy(protocol: str, identity: ABCActorIdentity, location: ActorLocation, proxy_type: str = None):
        proxy = ActorRegistrySingleton.get().get_proxy(protocol=protocol, actor_name=identity.get_name())
        if proxy is None:
            proxy = ProxyFactorySingleton.get().new_proxy(protocol=protocol, identity=identity, location=location,
                                                          proxy_type=proxy_type)
            ActorRegistrySingleton.get().register_proxy(proxy=proxy)
        return proxy



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import traceback
from typing import Tuple, List

from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.result_avro import ResultAvro
from fabric_mb.message_bus.messages.slice_avro import SliceAvro

from fabric_cf.actor.core.manage.error import Error
from fabric_cf.actor.core.manage.kafka.kafka_actor import KafkaActor
from fabric_cf.actor.core.util.id import ID


class ManageHelper:
    def __init__(self, *, logger):
        self.logger = logger

    def print_result(self, *, status: ResultAvro):
        self.logger.debug("Code={}".format(status.get_code()))
        if status.message is not None:
            self.logger.debug("Message={}".format(status.message))
        if status.details is not None:
            self.logger.debug("Details={}".format(status.details))

    @staticmethod
    def get_actor(*, actor_name: str) -> KafkaActor:
        from .kafka_processor import KafkaProcessorSingleton
        actor = KafkaProcessorSingleton.get().get_mgmt_actor(name=actor_name)
        return actor

    def do_get_slices(self, *, actor_name: str, callback_topic: str, slice_id: str = None,
                      id_token: str = None) -> Tuple[List[SliceAvro] or None, Error]:
        actor = self.get_actor(actor_name=actor_name)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(actor_name))
        try:
            actor.prepare(callback_topic=callback_topic)
            sid = None
            if slice_id is not None:
                sid = ID(uid=slice_id)
            return actor.get_slices(id_token=id_token, slice_id=sid), actor.get_last_error()
        except Exception:
            ex_str = traceback.format_exc()
            self.logger.error(ex_str)
        return None, actor.get_last_error()

    def do_get_reservations(self, *, actor_name: str, callback_topic: str, rid: str = None,
                            id_token: str = None) -> Tuple[List[ReservationMng] or None, Error]:
        actor = self.get_actor(actor_name=actor_name)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(actor_name))
        try:
            actor.prepare(callback_topic=callback_topic)
            reservation_id = None
            if rid is not None:
                reservation_id = ID(uid=rid)
            return actor.get_reservations(id_token=id_token, rid=reservation_id), actor.get_last_error()
        except Exception as e:
            ex_str = traceback.format_exc()
            self.logger.error(ex_str)
        return None, actor.get_last_error()

    def do_get_delegations(self, *, actor_name: str, callback_topic: str, did: str = None,
                           id_token: str = None) -> Tuple[List[DelegationAvro] or None, Error]:
        actor = self.get_actor(actor_name=actor_name)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(actor_name))
        try:
            actor.prepare(callback_topic=callback_topic)
            return actor.get_delegations(id_token=id_token, delegation_id=did), actor.get_last_error()
        except Exception as e:
            traceback.print_exc()
            ex_str = traceback.format_exc()
            self.logger.error(ex_str)
        return None, actor.get_last_error()

    def do_claim_delegations(self, *, broker: str, am_guid: ID, callback_topic: str, id_token: str = None,
                             did: str = None) -> Tuple[DelegationAvro or None, Error]:
        """
        Claim delegations by invoking Management Actor Claim Delegations API
        @param broker broker guid
        @param am_guid am guid
        @param callback_topic callback topic
        @param id_token id token
        @param did delegation id
        @return Tuple[Delegation, Error] Delegation on success and Error in case of failure
        """
        actor = self.get_actor(actor_name=broker)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(broker))
        try:
            actor.prepare(callback_topic=callback_topic)

            dlg = actor.claim_delegations(broker=am_guid, did=did, id_token=id_token)
            return dlg, actor.get_last_error()
        except Exception as e:
            self.logger.error(f"Exception occurred e: {e}")
            self.logger.error(traceback.format_exc())

        return None, actor.get_last_error()

    def claim_delegations(self, *, broker: str, am: str, callback_topic: str, did: str = None, id_token: str = None):
        """
        Claim delegations
        @param broker broker name
        @param am am name
        @param callback_topic callback topic
        @param id_token id token
        @param did delegation id
        """
        try:
            am_actor = self.get_actor(actor_name=am)
            broker_actor = self.get_actor(actor_name=broker)

            if am_actor is None or broker_actor is None:
                raise Exception("Invalid arguments am_actor {} or broker_actor {} not found".format(am_actor,
                                                                                                    broker_actor))

            delegations, error = self.do_get_delegations(actor_name=am, callback_topic=callback_topic, did=did,
                                                         id_token=id_token)
            if delegations is None:
                self.logger.debug("Error occurred while getting delegations for actor: {}".format(am))
                self.print_result(status=error.get_status())
                return

            if delegations is None or len(delegations) == 0:
                self.logger.debug("No delegations to be claimed from {} by {}:".format(am, broker))
                return

            claimed = False
            for d in delegations:
                if d.get_slice_object().get_slice_name() == broker:
                    self.logger.debug("Claiming Delegation# {}".format(d.get_delegation_id()))
                    delegation, error = self.do_claim_delegations(broker=broker, am_guid=am_actor.get_guid(),
                                                                  did=d.get_delegation_id(), callback_topic=callback_topic,
                                                                  id_token=id_token)
                    claimed = True
                    if delegation is not None:
                        self.logger.debug("Delegation claimed: {} ".format(delegation.get_delegation_id()))
                    else:
                        self.print_result(status=error.get_status())
            if not claimed:
                self.logger.error(f"No delegations found for Broker# {broker}")
        except Exception as e:
            self.logger.error(f"Exception occurred e: {e}")
            self.logger.error(traceback.format_exc())

    def do_reclaim_delegations(self, *, broker: str, am_guid: ID, callback_topic: str, id_token: str = None,
                               did: str = None) -> Tuple[DelegationAvro or None, Error]:
        """
        ReClaim delegations by invoking Management Actor ReClaim Delegations API
        @param broker broker guid
        @param am_guid am guid
        @param callback_topic callback topic
        @param id_token id token
        @param did delegation id
        @return Tuple[Delegation, Error] Delegation on success and Error in case of failure
        """
        actor = self.get_actor(actor_name=broker)

        if actor is None:
            raise Exception("Invalid arguments actor {} not found".format(broker))
        try:
            actor.prepare(callback_topic=callback_topic)

            dlg = actor.reclaim_delegations(broker=am_guid, did=did, id_token=id_token)
            return dlg, actor.get_last_error()
        except Exception as e:
            self.logger.error(f"Exception occurred e: {e}")
            self.logger.error(traceback.format_exc())

        return None, actor.get_last_error()

    def reclaim_delegations(self, *, broker: str, am: str, callback_topic: str, did: str = None,
                            id_token: str = None) -> bool:
        """
        ReClaim delegations
        @param broker broker name
        @param am am name
        @param callback_topic callback topic
        @param id_token id token
        @param did delegation id
        """
        try:
            am_actor = self.get_actor(actor_name=am)
            broker_actor = self.get_actor(actor_name=broker)

            if am_actor is None or broker_actor is None:
                raise Exception("Invalid arguments am_actor {} or broker_actor {} not found".format(am_actor,
                                                                                                    broker_actor))

            delegations, error = self.do_get_delegations(actor_name=am, callback_topic=callback_topic, did=did,
                                                         id_token=id_token)
            if delegations is None:
                self.logger.debug("Error occurred while getting delegations for actor: {}".format(am))
                self.print_result(status=error.get_status())
                return

            if delegations is None or len(delegations) == 0:
                self.logger.debug("No delegations to be reclaimed from {} by {}:".format(am, broker))
                return

            reclaim_failed = False
            for d in delegations:
                if d.get_slice_object().get_slice_name() == broker:
                    self.logger.debug("Reclaiming Delegation# {}".format(d.get_delegation_id()))
                    delegation, error = self.do_reclaim_delegations(broker=broker, am_guid=am_actor.get_guid(),
                                                                    did=d.get_delegation_id(),
                                                                    callback_topic=callback_topic,
                                                                    id_token=id_token)
                    if delegation is not None:
                        self.logger.debug("Delegation reclaimed: {} ".format(delegation.get_delegation_id()))
                    else:
                        reclaim_failed = True
                        self.print_result(status=error.get_status())
            return not reclaim_failed
        except Exception as e:
            self.logger.error(f"Exception occurred e: {e}")
            self.logger.error(traceback.format_exc())
        return False


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_base_plugin import ABCBasePlugin

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
    from fabric_cf.actor.core.core.unit import Unit
    from fabric_cf.actor.core.apis.abc_substrate_database import ABCSubstrateDatabase


class ABCSubstrate(ABCBasePlugin):
    """
    Base class for Substrate Plugin
    """
    @abstractmethod
    def get_substrate_database(self) -> ABCSubstrateDatabase:
        """
        Returns the substrate database.
        @return the substrate database
        """

    @abstractmethod
    def transfer_in(self, *, reservation: ABCReservationMixin, unit: Unit):
        """
        Transfers in a new unit into the substrate. Transfer in involves
        configuring the unit so that it is part of the substrate.
        @param reservation reservation that contains the unit
        @param unit unit to be transferred
        """

    @abstractmethod
    def transfer_out(self, *, reservation: ABCReservationMixin, unit: Unit):
        """
        Transfers an existing unit out of the substrate. Transfer out involves
        unconfiguring the unit, so that it is no longer part of the substrate.
        @param reservation reservation that contains the unit
        @param unit unit to be transferred
        """

    @abstractmethod
    def modify(self, *, reservation: ABCReservationMixin, unit: Unit):
        """
        Modifies an existing unit that is already part of the substrate.
        @param reservation reservation that contains the unit
        @param unit unit to be modified.
        """

    @abstractmethod
    def update_props(self, *, reservation: ABCReservationMixin, unit: Unit):
        """
        Updates only the properties of an existing unit that is already part of the substrate.
        @param reservation reservation that contains the unit
        @param unit unit to be modified.
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from typing import List

from fabric_mb.message_bus.messages.lease_reservation_avro import LeaseReservationAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.slice_avro import SliceAvro
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.network_node import NodeSliver
from fim.slivers.network_service import NetworkServiceSliver

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates, JoinState
from fabric_cf.actor.core.kernel.slice_state_machine import SliceState
from fabric_cf.actor.core.time.actor_clock import ActorClock


class ResponseBuilder:
    RESPONSE_STATUS = "status"
    STATUS_OK = "OK"
    STATUS_FAILURE = "FAILURE"

    RESPONSE_RESERVATIONS = "reservations"
    RESPONSE_SLICES = "slices"
    RESPONSE_MESSAGE = "message"
    RESPONSE_BQM = "bqm"
    RESPONSE_SLICE_MODEL = "slice_model"

    PROP_SLICE_ID = "slice_id"
    PROP_SLICE_NAME = "slice_name"
    PROP_SLICE_STATE = "slice_state"
    PROP_RESERVATION_ID = "reservation_id"
    PROP_RESOURCE_TYPE = "resource_type"
    PROP_RESERVATION_STATE = "reservation_state"
    PROP_LEASE_END_TIME = "lease_end"
    PROP_RESERVATION_PENDING_STATE = "pending_state"
    PROP_RESERVATION_JOIN_STATE = "join_state"
    PROP_NOTICES = "notices"
    PROP_GRAPH_NODE_ID = "graph_node_id"
    PROP_GRAPH_ID = "graph_id"
    PROP_NAME = "name"
    PROP_MANAGEMENT_IP = "management_ip"
    PROP_SITE = "site"
    PROP_CAPACITIES = "capacities"
    PROP_ALLOCATED_CAPACITIES = "allocated_capacities"
    PROP_LABELS = "labels"
    PROP_ALLOCATED_LABELS = "allocated_labels"
    PROP_CAPACITY_HINTS = "capacity_hints"

    @staticmethod
    def get_reservation_summary(*, res_list: List[ReservationMng], include_notices: bool = False,
                                include_sliver: bool = False) -> dict:
        """
        Get Reservation summary
        :param res_list:
        :param include_notices:
        :param include_sliver:
        :return:
        """
        reservations = []
        status = ResponseBuilder.STATUS_OK
        message = ""

        if res_list is not None:
            for reservation in res_list:
                res_dict = {ResponseBuilder.PROP_SLICE_ID: reservation.get_slice_id(),
                            ResponseBuilder.PROP_RESERVATION_ID: reservation.get_reservation_id(),
                            ResponseBuilder.PROP_RESERVATION_STATE: ReservationStates(reservation.get_state()).name}

                if reservation.get_pending_state() is not None:
                    res_dict[ResponseBuilder.PROP_RESERVATION_PENDING_STATE] = \
                        ReservationPendingStates(reservation.get_pending_state()).name

                if isinstance(reservation, LeaseReservationAvro) and reservation.get_join_state() is not None:
                    res_dict[ResponseBuilder.PROP_RESERVATION_JOIN_STATE] = JoinState(reservation.get_join_state()).name

                sliver = reservation.get_sliver()
                if sliver is not None:
                    res_dict[ResponseBuilder.PROP_GRAPH_NODE_ID] = sliver.node_id
                    if include_sliver:
                        res_dict = ResponseBuilder.get_sliver_json(sliver=sliver, result=res_dict)

                if include_notices:
                    res_dict[ResponseBuilder.PROP_NOTICES] = reservation.get_notices()

                if reservation.get_end() is not None:
                    end_time = ActorClock.from_milliseconds(milli_seconds=reservation.get_end())
                    res_dict[ResponseBuilder.PROP_LEASE_END_TIME] = end_time.strftime(Constants.RENEW_TIME_FORMAT)

                reservations.append(res_dict)
        else:
            status = ResponseBuilder.STATUS_FAILURE
            message = "No reservations were found/computed"

        response = {ResponseBuilder.RESPONSE_STATUS: status,
                    ResponseBuilder.RESPONSE_MESSAGE: message,
                    ResponseBuilder.RESPONSE_RESERVATIONS: reservations}

        return response

    @staticmethod
    def get_slice_summary(*, slice_list: List[SliceAvro], slice_id: str = None,
                          slice_states: List[SliceState] = None) -> dict:
        """
        Get slice summary
        :param slice_list:
        :param slice_id:
        :param slice_states:
        :return:
        """
        slices = []
        status = ResponseBuilder.STATUS_OK
        message = ""

        if slice_list is not None:
            for s in slice_list:
                slice_state = SliceState(s.get_state())
                if slice_id is None and slice_states is not None and slice_state not in slice_states:
                    continue
                s_dict = {ResponseBuilder.PROP_SLICE_ID: s.get_slice_id(),
                          ResponseBuilder.PROP_SLICE_NAME: s.get_slice_name(),
                          ResponseBuilder.PROP_GRAPH_ID: s.get_graph_id(),
                          ResponseBuilder.PROP_SLICE_STATE: slice_state.name,
                          }
                end_time = s.get_lease_end()
                if end_time is not None:
                    s_dict[ResponseBuilder.PROP_LEASE_END_TIME] = end_time.strftime(Constants.RENEW_TIME_FORMAT)
                slices.append(s_dict)
        else:
            status = ResponseBuilder.STATUS_FAILURE
            message = "No slices were found"

        response = {ResponseBuilder.RESPONSE_STATUS: status,
                    ResponseBuilder.RESPONSE_MESSAGE: message,
                    ResponseBuilder.RESPONSE_SLICES: slices}

        return response

    @staticmethod
    def get_broker_query_model_summary(*, bqm: str):
        """
        Get BQM
        :param bqm:
        :return:
        """
        status = ResponseBuilder.STATUS_OK
        message = ""

        if bqm is None:
            message = "No broker query model found"
            status = ResponseBuilder.STATUS_FAILURE

        response = {ResponseBuilder.RESPONSE_STATUS: status,
                    ResponseBuilder.RESPONSE_MESSAGE: message,
                    ResponseBuilder.RESPONSE_BQM: bqm}

        return response

    @staticmethod
    def get_slice_model_summary(*, slice_model: str):
        """
        Get Slice Model summary
        :param slice_model:
        :return:
        """
        status = ResponseBuilder.STATUS_OK
        message = ""

        if slice_model is None:
            message = "No slice model found"
            status = ResponseBuilder.STATUS_FAILURE

        response = {ResponseBuilder.RESPONSE_STATUS: status,
                    ResponseBuilder.RESPONSE_MESSAGE: message,
                    ResponseBuilder.RESPONSE_SLICE_MODEL: slice_model}

        return response

    @staticmethod
    def get_sliver_json(*, sliver: BaseSliver, result: dict) -> dict:
        result[ResponseBuilder.PROP_NAME] = sliver.get_name()
        result[ResponseBuilder.PROP_SITE] = sliver.get_site()
        result[ResponseBuilder.PROP_RESOURCE_TYPE] = str(sliver.get_type())
        if sliver.get_capacities() is not None:
            result[ResponseBuilder.PROP_CAPACITIES] = sliver.get_capacities().to_json()
        if sliver.get_capacity_allocations() is not None:
            result[ResponseBuilder.PROP_ALLOCATED_CAPACITIES] = sliver.get_capacity_allocations().to_json()
        if sliver.get_labels() is not None:
            result[ResponseBuilder.PROP_LABELS] = sliver.get_labels().to_json()
        if sliver.get_label_allocations() is not None:
            result[ResponseBuilder.PROP_ALLOCATED_LABELS] = sliver.get_label_allocations().to_json()

        # Network Node Specific Fields
        if isinstance(sliver, NodeSliver):
            if sliver.get_management_ip() is not None:
                result[ResponseBuilder.PROP_MANAGEMENT_IP] = sliver.get_management_ip()
            if sliver.get_capacity_hints() is not None:
                result[ResponseBuilder.PROP_CAPACITY_HINTS] = sliver.get_capacity_hints().to_json()

        # Network Service Specific Fields
        if isinstance(sliver, NetworkServiceSliver):
            print("TODO")

        return result

    @staticmethod
    def get_response_summary(*, rid_list: List[str]) -> dict:
        response = {}
        status = ResponseBuilder.STATUS_OK
        message = ""
        if rid_list is not None and len(rid_list) > 0:
            status = ResponseBuilder.STATUS_FAILURE
            response[ResponseBuilder.RESPONSE_RESERVATIONS] = rid_list

        response[ResponseBuilder.RESPONSE_STATUS] = status
        response[ResponseBuilder.RESPONSE_MESSAGE] = message

        return response


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import time
import unittest

from fim.slivers.base_sliver import BaseSliver
from fim.slivers.capacities_labels import Capacities, CapacityHints
from fim.slivers.instance_catalog import InstanceCatalog
from fim.slivers.network_node import NodeType, NodeSliver

from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
from fabric_cf.actor.core.apis.abc_authority import ABCAuthority
from fabric_cf.actor.core.apis.abc_authority_policy import ABCAuthorityPolicy
from fabric_cf.actor.core.apis.abc_authority_reservation import ABCAuthorityReservation
from fabric_cf.actor.core.apis.abc_database import ABCDatabase
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation, DelegationState
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import AuthorityException
from fabric_cf.actor.core.core.ticket import Ticket
from fabric_cf.actor.core.delegation.delegation_factory import DelegationFactory
from fabric_cf.actor.core.delegation.resource_ticket import ResourceTicketFactory
from fabric_cf.actor.core.kernel.authority_reservation import AuthorityReservationFactory
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.kernel.slice import SliceFactory
from fabric_cf.actor.core.plugins.handlers.ansible_handler_processor import AnsibleHandlerProcessor
from fabric_cf.actor.core.plugins.handlers.configuration_mapping import ConfigurationMapping
from fabric_cf.actor.core.plugins.substrate.db.substrate_actor_database import SubstrateActorDatabase
from fabric_cf.actor.core.plugins.substrate.substrate_mixin import SubstrateMixin
from fabric_cf.actor.core.policy.authority_calendar_policy import AuthorityCalendarPolicy
from fabric_cf.actor.core.policy.network_node_control import NetworkNodeControl
from fabric_cf.actor.core.policy.resource_control import ResourceControl
from fabric_cf.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.resource_type import ResourceType
from fabric_cf.actor.core.util.update_data import UpdateData
from fabric_cf.actor.security.auth_token import AuthToken
from fabric_cf.actor.test.base_test_case import BaseTestCase
from fabric_cf.actor.test.controller_callback_helper import ControllerCallbackHelper, IUpdateLeaseHandler
from fabric_cf.actor.test.fim_test_helper import FimTestHelper


class AuthorityCalendarPolicyTest(BaseTestCase, unittest.TestCase):
    DonateStartCycle = 10
    DonateEndCycle = 100
    TicketStartCycle = DonateStartCycle + 1
    TicketEndCycle = TicketStartCycle + 10
    TicketNewEndCycle = TicketEndCycle + 10
    TicketUnits = 1

    my_unit = None

    logger = logging.getLogger('AuthorityCalendarPolicyTest')
    log_format = '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s] - %(levelname)s - %(message)s'
    logging.basicConfig(format=log_format, filename="actor.log")
    logger.setLevel(logging.INFO)

    from fabric_cf.actor.core.container.globals import Globals
    Globals.config_file = "./config/config.test.yaml"
    Constants.SUPERBLOCK_LOCATION = './state_recovery.lock'

    from fabric_cf.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    FimTestHelper.n4j_imp.delete_all_graphs()
    arm, adm = FimTestHelper.generate_renci_adm()

    def build_sliver(self) -> BaseSliver:
        node_sliver = NodeSliver()
        node_sliver.resource_type = NodeType.VM
        node_sliver.node_id = "test-slice-node-1"
        cap = Capacities()
        cap.set_fields(core=4, ram=64, disk=500)
        catalog = InstanceCatalog()
        instance_type = catalog.map_capacities_to_instance(cap=cap)
        cap_hints = CapacityHints().set_fields(instance_type=instance_type)
        node_sliver.set_properties(name="node-1", type=NodeType.VM, site="RENC",
                                   capacities=cap, image_type='qcow2', image_ref='default_centos_8',
                                   capacity_hints=cap_hints)
        node_sliver.set_capacity_allocations(cap=catalog.get_instance_capacities(instance_type=instance_type))
        node_map = tuple([self.arm.graph_id, 'HX6VQ53'])
        node_sliver.set_node_map(node_map=node_map)
        return node_sliver

    def get_control(self) -> ResourceControl:
        control = NetworkNodeControl()
        control.add_type(rtype=ResourceType(resource_type=NodeType.VM.name))
        return control

    def get_delegation(self, actor: ABCActorMixin) -> ABCDelegation:
        slice_object = SliceFactory.create(slice_id=ID(), name="inventory-slice")
        slice_object.set_inventory(value=True)
        actor.register_slice(slice_object=slice_object)

        delegation_name = 'primary'

        dlg_obj = DelegationFactory.create(did=self.adm.get_graph_id(), slice_id=slice_object.get_slice_id(),
                                           delegation_name=delegation_name)
        dlg_obj.set_slice_object(slice_object=slice_object)
        dlg_obj.set_graph(graph=self.adm)
        actor.register_delegation(delegation=dlg_obj)
        return dlg_obj

    def check_before_donate_delegation(self, authority: ABCAuthority):
        policy = authority.get_policy()
        control = self.get_control()
        self.assertIsNotNone(control)
        self.assertEqual(len(policy.delegations), 0)

    def check_after_donate_donate_delegation(self, authority: ABCAuthority):
        policy = authority.get_policy()
        control = self.get_control()

        self.assertIsNotNone(control)
        self.assertEqual(len(policy.delegations), 1)

    def make_actor_database(self) -> ABCDatabase:
        db = SubstrateActorDatabase(user=self.db_user, password=self.db_pwd, database=self.db_name, db_host=self.db_host,
                                    logger=self.logger)
        return db

    def make_plugin(self):
        handler_processor = AnsibleHandlerProcessor()
        plugin = SubstrateMixin(actor=None, db=None, handler_processor=handler_processor)
        return plugin

    def get_config_map(self):
        config_map = ConfigurationMapping()
        config_map.set_key(key=NodeType.VM.name)
        config_map.set_class_name(class_name='NoOpHandler')
        config_map.set_module_name(module_name='fabric_cf.actor.handlers.no_op_handler')
        return config_map

    def get_authority(self, name: str = BaseTestCase.authority_name, guid: ID = BaseTestCase.authority_guid):
        db = self.get_container_database()
        db.reset_db()
        authority = super().get_authority()
        authority.set_recovered(value=True)
        authority.set_aggregate_resource_model(aggregate_resource_model=self.arm)
        authority.get_plugin().handler_processor.add_config_mapping(mapping=self.get_config_map())
        Term.set_clock(authority.get_actor_clock())
        return authority

    def get_authority_policy(self) -> ABCAuthorityPolicy:
        policy = AuthorityCalendarPolicy()
        policy.register_control(control=self.get_control())
        return policy

    def get_ticket(self, units: int, rtype: ResourceType, term: Term, source: ABCDelegation,
                   actor: ABCActorMixin) -> Ticket:
        resource_ticket = ResourceTicketFactory.create(issuer=self.broker_guid,
                                                       units=units,
                                                       term=term,
                                                       rtype=rtype)
        ticket = Ticket(resource_ticket=resource_ticket, plugin=actor.get_plugin(), authority=None)
        ticket.delegation_id = source.get_delegation_id()
        return ticket

    def get_request_slice(self):
        return SliceFactory.create(slice_id=ID(), name="test-slice")

    def get_request(self, term: Term, ticket: Ticket, sliver: BaseSliver):
        rset = ResourceSet(units=1, rtype=ResourceType(resource_type=sliver.resource_type.name), sliver=sliver)
        rset.set_resources(cset=ticket)
        slice_object = self.get_request_slice()
        return AuthorityReservationFactory.create(resources=rset, term=term, slice_obj=slice_object, rid=ID())

    def get_request_from_request(self, request: ABCAuthorityReservation, term: Term, ticket: Ticket):
        rset = ResourceSet(units=1, rtype=request.get_type())
        rset.set_resources(cset=ticket)
        result = AuthorityReservationFactory.create(resources=rset, term=term, slice_obj=request.get_slice(),
                                                    rid=request.get_reservation_id())
        result.set_sequence_in(sequence=request.get_sequence_in() + 1)
        return result

    def get_redeem_request(self, authority: ABCAuthority, delegation: ABCDelegation):
        req_start = authority.get_actor_clock().cycle_start_date(cycle=self.TicketStartCycle)
        req_end = authority.get_actor_clock().cycle_end_date(cycle=self.TicketEndCycle)
        req_term = Term(start=req_start, end=req_end)
        sliver = self.build_sliver()
        ticket = self.get_ticket(units=self.TicketUnits, rtype=ResourceType(resource_type=sliver.resource_type.name),
                                 term=req_term, source=delegation,
                                 actor=authority)

        request = self.get_request(req_term, ticket, sliver)
        return request

    def get_extend_lease_request(self, authority: ABCAuthority, delegation: ABCDelegation,
                                 request: ABCAuthorityReservation):
        req_start = authority.get_actor_clock().cycle_start_date(cycle=self.TicketStartCycle)
        req_new_start = authority.get_actor_clock().cycle_start_date(cycle=self.TicketEndCycle + 1)
        req_end = authority.get_actor_clock().cycle_end_date(cycle=self.TicketNewEndCycle)
        req_term = Term(start=req_start, end=req_end, new_start=req_new_start)
        rtype = ResourceType(resource_type=request.get_requested_resources().get_sliver().resource_type.name)
        ticket = self.get_ticket(self.TicketUnits, rtype, req_term, delegation, authority)
        new_request = self.get_request_from_request(request, req_term, ticket)
        return new_request

    def check_incoming_lease(self, request: ABCAuthorityReservation, incoming: ABCReservationMixin):
        self.assertIsNotNone(incoming)

        rset = incoming.get_resources()

        self.assertIsNotNone(rset)

        self.assertEqual(request.get_requested_units(), rset.get_units())

        self.assertEqual(incoming.get_term(), request.get_requested_term())

        uset = rset.get_resources()
        self.assertIsNotNone(uset)
        self.assertEqual(self.TicketUnits, uset.get_units())

        u = uset.get_set().values().__iter__().__next__()
        self.my_unit = u

    def check_incoming_close_lease(self, request: ABCAuthorityReservation, incoming: ABCReservationMixin):
        self.assertIsNotNone(incoming)

        rset = incoming.get_resources()

        self.assertIsNotNone(rset)

        self.assertEqual(request.get_requested_units(), rset.get_units())

        uset = rset.get_resources()
        self.assertIsNotNone(uset)
        self.assertEqual(0, uset.get_units())

    def check_incoming_extend_lease(self, request: ABCAuthorityReservation, incoming: ABCReservationMixin):
        self.assertIsNotNone(incoming)
        self.assertEqual(ReservationStates.Active, incoming.get_state())
        self.assertEqual(ReservationPendingStates.None_, incoming.get_pending_state())
        rset = incoming.get_resources()
        self.assertIsNotNone(rset)
        self.assertEqual(request.get_requested_units(), rset.get_units())
        self.assertEqual(incoming.get_term(), request.get_requested_term())

        uset = rset.get_resources()
        self.assertIsNotNone(uset)
        self.assertEqual(self.TicketUnits, uset.get_units())

        u = uset.get_set().values().__iter__().__next__()
        self.assertEqual(self.my_unit, u)

    def test_a_create(self):
        authority = self.get_authority()
        self.assertIsNotNone(authority)

    def test_b_donate(self):
        site = self.get_authority()
        policy = site.get_policy()
        self.check_before_donate_delegation(site)
        delegation = self.get_delegation(site)
        policy.donate_delegation(delegation=delegation)
        self.check_after_donate_donate_delegation(site)

    def test_c_redeem(self):
        site = self.get_authority()
        policy = site.get_policy()
        controller = self.get_controller()
        proxy = ControllerCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        authority_proxy = ControllerCallbackHelper(name=site.get_name(), guid=site.get_guid())

        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=authority_proxy)

        delegation = self.get_delegation(site)
        policy.donate_delegation(delegation=delegation)

        request = self.get_redeem_request(site, delegation)

        class UpdateLeaseHandler(IUpdateLeaseHandler):
            def __init__(self, parent, request):
                self.waiting_for_lease = True
                self.waiting_for_close = False
                self.parent = parent
                self.request = request

            def handle_update_lease(self, reservation: ABCReservationMixin, update_data: UpdateData, caller: AuthToken):
                if self.waiting_for_lease:
                    self.parent.check_incoming_lease(self.request, reservation)
                    self.waiting_for_lease = False
                    self.waiting_for_close = True
                elif self.waiting_for_close:
                    self.parent.assertTrue(site.get_current_cycle() >= AuthorityCalendarPolicyTest.TicketEndCycle)
                    self.parent.check_incoming_close_lease(self.request, reservation)
                    self.waiting_for_close = False
                else:
                    raise AuthorityException(Constants.INVALID_STATE)

            def check_termination(self):
                self.parent.assertFalse(self.waiting_for_lease)
                self.parent.assertFalse(self.waiting_for_close)

        handler = UpdateLeaseHandler(self, request)
        proxy.set_update_lease_handler(handler=handler)

        cycle = 1
        site.external_tick(cycle=cycle)
        cycle += 1

        print("Redeeming request...")
        # Force delegation to be in Delegated State
        delegation.state = DelegationState.Delegated
        site.redeem(reservation=request, callback=proxy, caller=proxy.get_identity())

        for c in range(cycle, self.DonateEndCycle):
            site.external_tick(cycle=c)
            while site.get_current_cycle() != c:
                time.sleep(0.001)

        handler.check_termination()

    def test_d_extend_lease(self):
        site = self.get_authority()
        policy = site.get_policy()
        controller = self.get_controller()
        proxy = ControllerCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        auth_proxy = ControllerCallbackHelper(name=site.get_name(), guid=site.get_guid())
        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=auth_proxy)

        delegation = self.get_delegation(site)
        policy.donate_delegation(delegation=delegation)

        request = self.get_redeem_request(site, delegation)

        extend = self.get_extend_lease_request(site, delegation, request)

        class UpdateLeaseHandler(IUpdateLeaseHandler):
            def __init__(self, parent, request):
                self.waiting_for_extend_lease = False
                self.waiting_for_lease = True
                self.waiting_for_close = False
                self.parent = parent
                self.request = request

            def handle_update_lease(self, reservation: ABCReservationMixin, update_data: UpdateData, caller: AuthToken):
                if self.waiting_for_lease:
                    self.parent.check_incoming_lease(self.request, reservation)
                    self.waiting_for_lease = False
                    self.waiting_for_close = True
                elif self.waiting_for_close:
                    self.parent.assertTrue(site.get_current_cycle() >= AuthorityCalendarPolicyTest.TicketEndCycle)
                    self.parent.check_incoming_close_lease(self.request, reservation)
                    self.waiting_for_close = False
                else:
                    raise AuthorityException(Constants.INVALID_STATE)

            def check_termination(self):
                self.parent.assertFalse(self.waiting_for_lease)
                self.parent.assertFalse(self.waiting_for_close)

        handler = UpdateLeaseHandler(self, request)

        proxy.set_update_lease_handler(handler=handler)
        cycle = 1
        site.external_tick(cycle=cycle)
        cycle += 1
        print("Redeeming request")
        # Force delegation to be in Delegated State
        delegation.state = DelegationState.Delegated

        site.redeem(reservation=request, callback=proxy, caller=proxy.get_identity())
        for cycle in range(cycle, self.DonateEndCycle):
            if cycle == self.TicketEndCycle - 50:
                print("Extending Lease")
                site.extend_lease(reservation=extend, caller=proxy.get_identity())
            site.external_tick(cycle=cycle)
            while site.get_current_cycle() != cycle:
                time.sleep(0.001)
        handler.check_termination()

    def test_e_close(self):
        site = self.get_authority()
        policy = site.get_policy()
        controller = self.get_controller()
        proxy = ControllerCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        authority_proxy = ControllerCallbackHelper(name=site.get_name(), guid=site.get_guid())

        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=authority_proxy)

        delegation = self.get_delegation(site)
        policy.donate_delegation(delegation=delegation)

        request = self.get_redeem_request(site, delegation)

        class UpdateLeaseHandler(IUpdateLeaseHandler):
            def __init__(self, parent):
                self.waiting_for_lease = True
                self.waiting_for_close = False
                self.parent = parent

            def handle_update_lease(self, reservation: ABCReservationMixin, update_data: UpdateData, caller: AuthToken):
                if self.waiting_for_lease:
                    self.parent.check_incoming_lease(request=request, incoming=reservation)
                    self.waiting_for_lease = False
                    self.waiting_for_close = True
                elif self.waiting_for_close:
                    print(site.get_current_cycle())
                    self.parent.assertTrue(site.get_current_cycle() < AuthorityCalendarPolicyTest.TicketEndCycle)
                    self.parent.check_incoming_close_lease(request=request, incoming=reservation)
                    self.waiting_for_close = False
                else:
                    raise AuthorityException(Constants.INVALID_STATE)

            def check_termination(self):
                self.parent.assertFalse(self.waiting_for_lease)
                self.parent.assertFalse(self.waiting_for_close)

        handler = UpdateLeaseHandler(self)
        proxy.set_update_lease_handler(handler=handler)

        cycle = 1
        site.external_tick(cycle=cycle)
        cycle += 1
        print("Redeeming request...")
        # Force delegation to be in Delegated State
        delegation.state = DelegationState.Delegated
        site.redeem(reservation=request, callback=proxy, caller=proxy.get_identity())

        for cycle in range(cycle, self.DonateEndCycle):
            if cycle == self.TicketEndCycle - 5:
                site.close(reservation=request)
            site.external_tick(cycle=cycle)
            while site.get_current_cycle() != cycle:
                time.sleep(0.001)

        handler.check_termination()


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from typing import List

from fabric_cf.actor.core.util.id import ID
from fabric_cf.orchestrator.core.i_status_update_callback import IStatusUpdateCallback
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng


class ReservationDependencyStatusUpdate(IStatusUpdateCallback):
    def __init__(self):
        self.reservation = None

    def success(self, *, ok: List[ID], act_on: List[ID]):
        """
        Success callback
        :param ok:
        :param act_on:
        :return:
        """
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        logger = GlobalsSingleton.get().get_logger()
        # TODO

        logger.debug("Success")

    def failure(self, *, failed: List[ID], ok: List[ID], act_on: List[ID]):
        """
        Failure callback
        :param failed:
        :param ok:
        :param act_on:
        :return:
        """
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        logger = GlobalsSingleton.get().get_logger()

        logger.debug("Failure")
        # TODO

    def get_reservation(self) -> ReservationMng:
        """
        Get Reservation
        :return:
        """
        return self.reservation

    def set_reservation(self, *, reservation: ReservationMng):
        """
        Set reservation
        :param reservation:
        :return:
        """
        self.reservation = reservation



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING, List

from fabric_mb.message_bus.messages.delegation_avro import DelegationAvro
from fabric_mb.message_bus.messages.broker_query_model_avro import BrokerQueryModelAvro
from fim.user import GraphFormat

from fabric_cf.actor.core.apis.abc_component import ABCComponent


if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
    from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
    from fabric_mb.message_bus.messages.proxy_avro import ProxyAvro
    from fabric_cf.actor.core.util.id import ID


class ABCMgmtClientActor(ABCComponent):
    """
    Implements base class for Management Interface for a Client Actor
    """
    @abstractmethod
    def add_reservation(self, *, reservation: TicketReservationAvro) -> ID:
        """
        Adds the reservation to the actor's state and returns the assigned reservation ID.
        The reservation must refer to a valid slice.
        The reservation ID is also attached to the passed in reservation object.
        @param reservation reservation
        @return null on failure, assigned reservation ID otherwise
        """

    @abstractmethod
    def add_reservations(self, *, reservations: List[ReservationMng])->list:
        """
        Adds all reservations to the actor's state and returns the assigned reservation ID.
        Each reservation must refer to a valid slice.
        The reservation ID is also attached to the passed in reservation object.
        The operation is atomic: all of the reservations are added or none of them is added.
        @param reservations reservation
        @return null on failure, list of assigned ReservationIDs on success.
        """

    @abstractmethod
    def demand_reservation_rid(self, *, rid: ID) -> bool:
        """
        Demands the specified reservation.
        A reservation can be demanded only if has been added and it is in the Nascent state.
        @param rid reservation id
        @return true for success; false otherwise
        """

    @abstractmethod
    def demand_reservation(self, *, reservation: ReservationMng) -> bool:
        """
        Updates the reservation and issues a demand for it.
        A reservation can be demanded only if has been added and it is in the Nascent state.
        The reservation must refer to a valid slice. It can also indicate
        redeem predecessors.
        @param reservation reservation
        @return true for success; false otherwise
        """

    @abstractmethod
    def get_brokers(self, *, broker: ID = None, id_token: str = None) -> List[ProxyAvro]:
        """
        Retuns all brokers known to the actor.
        @param broker broker id
        @param id_token identity token generated by Credmgr
        @return list of all brokers
        """

    @abstractmethod
    def add_broker(self, *, broker: ProxyAvro) -> bool:
        """
        Adds a new broker.
        @param broker broker
        @return true for sucess; false otherwise
        """

    @abstractmethod
    def get_broker_query_model(self, *, broker: ID, id_token: str, level: int,
                               graph_format: GraphFormat) -> BrokerQueryModelAvro:
        """
        Obtains the resources available at the specified broker
        @param broker broker
        @param id_token identity token generated by Credmgr
        @param level: level of details
        @param graph_format: Graph Format
        @return BQM
        """

    @abstractmethod
    def extend_reservation(self, *, reservation: ID, new_end_time: datetime) -> bool:
        """
        Extend a reservation
        @params reservation: reservation id
        @params new_end_time: new end time
        @return true for success and false for failure
        """

    @abstractmethod
    def claim_delegations(self, *, broker: ID, did: ID, id_token: str = None) -> DelegationAvro:
        """
        Claims delegations exported by the specified broker
        @param broker broker guid
        @param did reservation id
        @param id_token id token
        @return reservation
        """

    @abstractmethod
    def reclaim_delegations(self, *, broker: ID, did: ID, id_token: str = None) -> DelegationAvro:
        """
        Reclaim delegations exported by the specified broker
        @param broker broker guid
        @param did reservation id
        @param id_token id token
        @return reservation
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List, Tuple

from fim.slivers.base_sliver import BaseSliver

from fabric_cf.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin


class InventoryForType:
    @abstractmethod
    def free(self, *, count: int, request: dict = None, resource: dict = None) -> dict:
        """
        Frees the specified number of resource units.
        @param count number of units
        @param request request properties
        @param resource resource properties
        @return new resource properties
        """

    @abstractmethod
    def allocate(self, *, rid: ID, requested_sliver: BaseSliver, graph_id: str, graph_node: BaseSliver,
                 existing_reservations: List[ABCReservationMixin]) -> Tuple[str, BaseSliver]:
        """
        Allocate an extending or ticketing reservation
        :param rid: reservation to be allocated
        :param requested_sliver: requested sliver
        :param graph_id: BQM graph id
        :param graph_node: BQM graph node identified to serve the reservation
        :param existing_reservations: Existing Reservations served by the same BQM node
        :return: Tuple of Delegation Id and the Requested Sliver annotated with BQM Node Id and other properties
        :raises: BrokerException in case the request cannot be satisfied
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import datetime

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import TimeException
from fabric_cf.actor.core.time.actor_clock import ActorClock


class Term:
    """
    Represents the term for a ticket or lease (i.e., a "reservation"). A term
    specifies the range of time units during which the term is valid.

    A term consists of start and end time. The interval represented by a term is
    closed on both ends. The term's newStart field represents the
    start time of the latest term extension. For extended terms
    start is constant, while newStart and end change to reflect the extensions.

    The length of a term is measured as the number of milliseconds in the
    closed interval [newStart, end].
    This number will be returned when calling getLength(). To obtain the
    number of milliseconds in the closed interval [start, end], use getFullLength().
    """

    # Flag that controls, whether cycle numbers should be calculated.
    set_cycles: bool = True
    clock: ActorClock = None
    HH_MM_TIME_FORMAT = "%Y-%m-%d %H:%M"

    @staticmethod
    def set_clock(clock: ActorClock):
        """
        Sets the clock to use when computing cycles.
        @params clock : clock instance
        """
        Term.clock = clock

    @staticmethod
    def delta(d1: datetime, d2: datetime) -> int:
        """
        Computes the difference in microseconds between two dates.
        @params d1 : date 1
        @params d2: date 2
        @returns the difference in microseconds between two dates.
        """
        d2_ms = ActorClock.to_milliseconds(when=d2)
        d1_ms = ActorClock.to_milliseconds(when=d1)

        return d2_ms - d1_ms

    @staticmethod
    def get_readable_date(date: datetime) -> str:
        """
        Returns a readable date.
        @params date: datetime object
        @returns Returns a readable date.
        """
        return date.strftime("%Y-%m-%d %H:%M:%S.%f %Z")

    @staticmethod
    def _get_clock():
        """
        Returns the clock factory in use in this container.
        @returns clock factory
        """
        if Term.clock is None:
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            actor_container = GlobalsSingleton.get().get_container()
            if actor_container is not None:
                Term.clock = actor_container.get_actor_clock()

            if Term.clock is None:
                raise RuntimeError("Failed to obtain the actor clock from the container")

        return Term.clock

    def __init__(self, *, start: datetime = None, end: datetime = None, new_start: datetime = None, length: int = None):
        """
        Creates a new term.
        @params start: start time
        @params end: end time
        @params new_start: new start time
        @params length: length in ms
        """
        # Start time: first valid millisecond.
        if start is not None:
            self.start_time = start
        else:
            self.start_time = datetime.utcnow()
        # End time: last valid millisecond.
        if end is not None:
            self.end_time = end
        else:
            if start is not None and length is not None and length > 1:
                start_ms = ActorClock.to_milliseconds(when=start) + length - 1
                self.end_time = ActorClock.from_milliseconds(milli_seconds=start_ms)
            else:
                raise TimeException("Invalid arguments, length and end both not specified")

        # Start time for this section of the lease
        if new_start is not None:
            self.new_start_time = new_start
        else:
            if start is not None:
                self.new_start_time = start
            else:
                self.new_start_time = datetime.utcnow()
        # Start cycle. Used only for debugging.
        self.cycle_start = 0
        # End cycle. Used only for debugging.
        self.cycle_end = 0
        # New start cycle. Used only for debugging.
        self.cycle_new_start = 0
        self._set_cycles()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['cycle_start']
        del state['cycle_end']
        del state['cycle_new_start']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.cycle_start = 0
        self.cycle_end = 0
        self.cycle_new_start = 0

    def _set_cycles(self):
        """
        Set cycles
        """
        if Term.set_cycles:
            clock = self._get_clock()
            if self.start_time is not None:
                self.cycle_start = clock.cycle(when=self.start_time)
            if self.end_time is not None:
                self.cycle_end = clock.cycle(when=self.end_time)
            if self.new_start_time is not None:
                self.cycle_new_start = clock.cycle(when=self.new_start_time)

    def change_length(self, *, length: int):
        """
        Creates a new term from the term. The new term has the same start time
        but different length.
        @param length : new term length (milliseconds)
        @returns term starting at the same time as this term but with the specified length
        """
        return Term(start=self.start_time, length=length)

    def contains(self, *, date: datetime = None, term=None) -> bool:
        """
        Checks if the term contains the given date or term.
        @params date: the date to check
        @params term: Term to check
        @returns true if the term contains the given date or term; false otherwise
        """
        if date is None and term is None:
            raise TimeException(Constants.INVALID_ARGUMENT)

        if self.start_time is None or self.end_time is None:
            raise TimeException(Constants.INVALID_STATE)

        if date is not None:
            return (not self.start_time > date) and (not self.end_time < date)

        if term is not None:
            if not isinstance(term, Term):
                raise TimeException(Constants.INVALID_ARGUMENT)
            return (not self.start_time > term.start_time) and (not self.end_time < term.end_time)

        return False

    def ends_after(self, *, date: datetime) -> bool:
        """
        Checks if the term ends after the given date.
        @params date : date to check against
        @returns true if the term ends after the given date
        """
        if date is None:
            raise TimeException(Constants.INVALID_ARGUMENT)

        if self.end_time is None:
            raise TimeException(Constants.INVALID_STATE)

        return date < self.end_time

    def ends_before(self, *, date: datetime) -> bool:
        """
        Checks if the term ends after the given date.
        @params date : date to check against
        @returns true if the term ends before the given date
        """
        if date is None:
            raise TimeException(Constants.INVALID_ARGUMENT)

        if self.end_time is None:
            raise TimeException(Constants.INVALID_STATE)

        return date > self.end_time

    def enforce_extends_term(self, *, old_term):
        """
        Checks if this term extends the old one. In case this term does not
        extend old term, logs the error and throws an exception.
        @params old_term : old term
        @raises Exception if this term does not extend the old term
        """
        if not isinstance(old_term, Term):
            raise TimeException("Invalid type: {}".format(type(old_term)))

        flag = self.extends_term(old_term=old_term)
        if flag is False:
            raise TimeException("New term does not extend previous term")

    def __eq__(self, other):
        """
        Compares two terms.
        @params other other term to compare to
        @returns true if both terms are equal
        """
        if other is None:
            return False
        if not isinstance(other, Term):
            return NotImplemented

        if self.start_time is not None and self.start_time != other.start_time:
            return False

        if self.end_time is not None and self.end_time != other.end_time:
            return False

        if self.new_start_time is not None and self.new_start_time != other.new_start_time:
            return False

        return True

    def expired(self, *, date: datetime) -> bool:
        """
        Checks if the term's expiration date is before the specified time.
        @params time : the time to check against
        @returns true if the term expires before the specified time
        """
        if date is None:
            raise TimeException(Constants.INVALID_ARGUMENT)

        if self.end_time is None:
            raise TimeException(Constants.INVALID_STATE)

        return date > self.end_time

    def extend(self, *, length: int = 0):
        """
        Creates a new term as an extension of the specified term. The term is
        extended with the current term length.
        @params length new term length in milliseconds
        @returns term extended with the current term length
        """
        if self.start_time is None or self.end_time is None:
            raise TimeException(Constants.INVALID_STATE)
        length_to_use = self.get_length()
        if length != 0:
            length_to_use = length

        new_start_ms = ActorClock.to_milliseconds(when=self.end_time) + 1
        new_start = ActorClock.from_milliseconds(milli_seconds=new_start_ms)
        end = ActorClock.from_milliseconds(milli_seconds=new_start_ms + length_to_use - 1)

        return Term(start=self.start_time, end=end, new_start=new_start)

    def extends_term(self, *, old_term) -> bool:
        """
        Checks if this term extends the old term.
        @params old_term: old term to check against
        @returns true if this term extends the old one, false if not
        """
        if old_term is None or old_term.start_time is None or old_term.end_time is None:
            raise TimeException(Constants.INVALID_STATE)

        if self.start_time is None or self.end_time is None or self.new_start_time is None:
            raise TimeException(Constants.INVALID_STATE)

        start_time_hh_mm = self.start_time.strftime(self.HH_MM_TIME_FORMAT)
        old_start_time_hh_mm = old_term.start_time.strftime(self.HH_MM_TIME_FORMAT)
        return start_time_hh_mm == old_start_time_hh_mm and self.end_time > self.new_start_time

    def get_end_time(self) -> datetime:
        """
        Returns the end time
        @returns end time
        """
        return self.end_time

    def get_full_length(self) -> int:
        """
        Returns the full length of a term in milliseconds. The full length of a
        term is the number of milliseconds in the closed interval [start, end].
        @returns term length
        """
        if self.start_time is None or self.end_time is None:
            raise TimeException(Constants.INVALID_STATE)

        start_ms = ActorClock.to_milliseconds(when=self.start_time)
        end_ms = ActorClock.to_milliseconds(when=self.end_time)

        return end_ms - start_ms + 1

    def get_length(self) -> int:
        """
        Returns the length of a term in milliseconds. The length of a term is the
        number of milliseconds in the closed interval [new_start, end]
        @returns term length
        """
        if self.new_start_time is None or self.end_time is None:
            raise TimeException(Constants.INVALID_STATE)

        new_start_ms = ActorClock.to_milliseconds(when=self.new_start_time)
        end_ms = ActorClock.to_milliseconds(when=self.end_time)

        return end_ms - new_start_ms + 1

    def get_new_start_time(self) -> datetime:
        """
        returns new start time
        @returns new start time
        """
        return self.new_start_time

    def get_start_time(self) -> datetime:
        """
        Returns start time
        @returns start time
        """
        return self.start_time

    def set_end_time(self, *, date: datetime):
        """
        Set end time
        """
        self.end_time = date
        self._set_cycles()

    def set_new_start_time(self, *, date: datetime):
        """
        Set new_start time
        """
        self.new_start_time = date
        self._set_cycles()

    def set_start_time(self, *, date: datetime):
        """
        Set start time
        """
        self.start_time = date
        self._set_cycles()

    def shift(self, *, date: datetime):
        """
        Creates a new term from the term. The new term is shifted in time to
        start at the specified start time.
        @params date : start time
        @returns term starting at the specified time, with the length of the current term
        """
        if date is None:
            raise TimeException("Invalid argument")

        return Term(start=date, length=self.get_length())

    def __str__(self):
        if Term.set_cycles:
            return "term=[{}:{}:{}]".format(self.cycle_start, self.cycle_new_start, self.cycle_end)
        else:
            return "Start: {} End: {}".format(Term.get_readable_date(self.start_time),
                                              Term.get_readable_date(self.end_time))

    def validate(self):
        """
        Validates the term
        @raises Exception thrown if invalid start time for term thrown if invalid
                end time for term thrown if negative duration for term
        """
        if self.start_time is None:
            raise TimeException("Invalid start time")

        if self.end_time is None:
            raise TimeException("Invalid end time")

        if self.end_time < self.start_time:
            raise TimeException("negative duration for term")

        if self.start_time == self.end_time:
            raise TimeException("zero duration for term")

        if self.new_start_time < self.start_time:
            raise TimeException("new start before start")

        if self.new_start_time > self.end_time:
            raise TimeException("new start after end")

    def clone(self):
        return Term(start=self.start_time, new_start=self.new_start_time, end=self.end_time)



from typing import List

from fabric_cf.orchestrator.swagger_server.response import authorization_controller as rc

"""
controller generated to handled auth operation described at:
https://connexion.readthedocs.io/en/latest/security.html
"""
def check_bearerAuth(token):
    return rc.check_bearerAuth(token=token)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING
from fabric_cf.actor.core.apis.abc_reservation_mixin import ReservationCategory
from fabric_cf.actor.core.apis.abc_kernel_authority_reservation_mixin import ABCKernelAuthorityReservationMixin
from fabric_cf.actor.core.common.exceptions import AuthorityException
from fabric_cf.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric_cf.actor.core.kernel.request_types import RequestTypes
from fabric_cf.actor.core.kernel.reservation_server import ReservationServer
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric_cf.actor.core.util.rpc_exception import RPCError

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.apis.abc_callback_proxy import ABCCallbackProxy
    from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice
    from fabric_cf.actor.core.kernel.failed_rpc import FailedRPC
    from fabric_cf.actor.core.apis.abc_kernel_slice import ABCKernelSlice
    from fabric_cf.actor.core.kernel.resource_set import ResourceSet
    from fabric_cf.actor.core.time.term import Term
    from fabric_cf.actor.core.util.id import ID


class AuthorityReservation(ReservationServer, ABCKernelAuthorityReservationMixin):
    """
    AuthorityReservation controls the state machine for a reservation on the
    authority side. It coordinates resource allocation, lease generation,
    priming, and shutdown of reservations.
    """
    UnexpectedExceptionString = "Unexpected reservation state: state={} pending={}"

    def __init__(self, *, rid: ID, resources: ResourceSet, term: Term, slice_object: ABCKernelSlice):
        super().__init__(rid=rid, resources=resources, term=term, slice_object=slice_object)
        # The ticket.
        self.ticket = None
        # Policies use this flag to instruct the core to send reservations to the client even if they have deficit.
        self.send_with_deficit = True
        # True if we notified the client about the fact that the reservation had failed
        self.notified_about_failure = False
        # Creates a new "blank" reservation instance. Used during recovery.
        self.category = ReservationCategory.Authority
        self.broker_callback = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['actor']
        del state['logger']
        del state['slice']
        del state['approved']
        del state['previous_resources']
        del state['bid_pending']
        del state['dirty']
        del state['expired']
        del state['pending_recover']
        del state['state_transition']
        del state['service_pending']

        del state['policy']

        del state['notified_about_failure']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.actor = None
        self.logger = None
        self.slice = None
        self.approved = False
        self.previous_resources = None
        self.bid_pending = False
        self.dirty = False
        self.expired = False
        self.pending_recover = False
        self.state_transition = False
        self.service_pending = ReservationPendingStates.None_

        self.policy = None

        self.notified_about_failure = False

    def restore(self, *, actor: ABCActorMixin, slice_obj: ABCSlice):
        """
        Must be invoked after creating reservation from unpickling
        """
        super().restore(actor=actor, slice_obj=slice_obj)
        self.notified_about_failure = False

    def set_broker_callback(self, *, broker_callback: ABCCallbackProxy):
        self.broker_callback = broker_callback

    def get_broker_callback(self) -> ABCCallbackProxy:
        return self.broker_callback

    def prepare(self, *, callback: ABCCallbackProxy, logger):
        self.set_logger(logger=logger)
        self.callback = callback
        self.requested_resources.validate_incoming_ticket(term=self.requested_term)

        if self.rid is None:
            self.error(err="no reservation ID specified for request")

        self.state = ReservationStates.Ticketed

    def reserve(self, *, policy: ABCPolicy):
        self.nothing_pending()
        self.incoming_request()
        if self.is_active():
            self.error(err="reservation already holds a lease")

        self.policy = policy
        self.approved = False
        self.bid_pending = True
        self.pending_recover = False
        self.map_and_update(extend=False)

    def service_reserve(self):
        try:
            if self.resources is not None:
                self.resources.service_reserve_site()
        except Exception as e:
            self.logger.error("authority failed servicing reserve", exception=e)
            self.fail_notify(message=str(e))

    def extend_lease(self):
        self.nothing_pending()
        self.incoming_request()

        if not self.is_active():
            self.error(err="reservation does not yet hold a lease")

        if not self.requested_term.extends_term(old_term=self.term):
            self.error(err="requested term does not extend current term for extendLease")

        self.approved = False
        self.bid_pending = True
        self.pending_recover = False
        self.map_and_update(extend=True)

    def modify_lease(self):
        self.nothing_pending()
        self.incoming_request()

        if not self.is_active():
            self.error(err="reservation does not yet hold a lease")

        self.approved = True
        self.bid_pending = True
        self.pending_recover = False
        self.map_and_update_modify_lease()

    def service_extend_lease(self):
        assert (self.state == ReservationStates.Failed and self.pending_state == ReservationPendingStates.None_) or \
                self.pending_state == ReservationPendingStates.ExtendingLease or \
                self.pending_state == ReservationPendingStates.Priming

        try:
            if self.pending_state == ReservationPendingStates.Priming:
                self.resources.service_extend()
        except Exception as e:
            self.logger.error("authority failed servicing extendLease e: {}".format(e))
            self.fail_notify(message=str(e))

    def service_modify_lease(self):
        assert (self.state == ReservationStates.Failed and self.pending_state == ReservationPendingStates.None_) or \
               self.pending_state == ReservationPendingStates.ModifyingLease or \
               self.pending_state == ReservationPendingStates.Priming

        try:
            if self.pending_state == ReservationPendingStates.Priming:
                self.resources.service_modify()
        except Exception as e:
            self.logger.error("authority failed servicing modifylease e: {}".format(e))
            self.fail_notify(message=str(e))

    def close(self):
        self.logger.debug("Processing  close for #{}".format(self.rid))
        self.transition(prefix="external close", state=self.state, pending=ReservationPendingStates.Closing)

    def service_close(self):
        if self.resources is not None:
            self.resources.close()

    def handle_duplicate_request(self, *, operation: RequestTypes):
        # The general idea is to do nothing if we are in the process of
        # performing a pending operation or about to reissue a
        # ticket/extendTicket after recovery. If there is nothing pending for
        # this reservation, we resend the last update.

        if operation == RequestTypes.RequestRedeem or operation == RequestTypes.RequestExtendLease or \
                operation == RequestTypes.RequestModifyLease:
            if self.pending_state == ReservationPendingStates.None_ and not self.bid_pending and \
                    not self.pending_recover:
                self.generate_update()
        else:
            raise AuthorityException("Unsupported operation: {}".format(operation))

    def map_and_update(self, *, extend: bool) -> bool:
        """
        Calls the policy to fill a request, with associated state transitions.

        @param extend
                   true if this request is an extend
        @return boolean success
        """
        success = False
        granted = False
        if self.state == ReservationStates.Failed:
            # Must be a previous failure, or policy marked it as failed. Send update to reset client.
            self.generate_update()
        elif self.state == ReservationStates.Ticketed:
            assert not extend
            try:
                self.transition(prefix="redeeming", state=ReservationStates.Ticketed,
                                pending=ReservationPendingStates.Redeeming)
                # If the policy has processed this reservation, set granted to
                # true so that we can start priming the resources. If the
                # policy has not yet processed this reservation (binPending is
                # true) then call the policy. The policy may choose to process
                # the request immediately (true) or to defer it (false). In
                # case of a deferred request, we will eventually come back to
                # this method after the policy has done its job.
                if self.is_bid_pending():
                    granted = self.policy.bind(reservation=self)
                else:
                    granted = True
            except Exception as e:
                self.logger.error("authority policy bind e: {}".format(e))
                self.fail_notify(message=str(e))

            if granted:
                try:
                    success = True
                    self.ticket = self.requested_resources
                    self.term = self.approved_term
                    self.resources = self.requested_resources.abstract_clone()
                    self.resources.units = 0
                    self.resources.update(reservation=self, resource_set=self.approved_resources)
                    self.transition(prefix="redeem", state=ReservationStates.Ticketed,
                                    pending=ReservationPendingStates.Priming)
                except Exception as e:
                    self.logger.error("authority redeem e: {}".format(e))
                    self.logger.error(traceback.format_exc())
                    self.fail_notify(message=str(e))
        elif self.state == ReservationStates.Active:
            assert extend
            try:
                self.transition(prefix="extending lease", state=ReservationStates.Active,
                                pending=ReservationPendingStates.ExtendingLease)
                # If the policy has processed this reservation, set granted to
                # true so that we can start priming the resources. If the
                # policy has not yet processed this reservation (binPending is
                # true) then call the policy. The policy may choose to process
                # the request immediately (true) or to defer it (false). In
                # case of a deferred request, we will eventually come back to
                # this method after the policy has done its job.
                if self.is_bid_pending():
                    granted = self.policy.extend_authority(reservation=self)
                else:
                    granted = True

                if granted:
                    success = True
                    self.extended = True
                    self.previous_term = self.term
                    self.ticket = self.requested_resources
                    self.term = self.approved_term

                    self.resources.update(reservation=self, resource_set=self.approved_resources)
                    self.transition(prefix="extend lease", state=ReservationStates.Active,
                                    pending=ReservationPendingStates.Priming)
            except Exception as e:
                self.logger.error("authority mapper extend e: {}".format(e))
                self.logger.error(traceback.format_exc())
                self.fail_notify(message=str(e))
        else:
            self.fail(message="mapAndUpdate: unexpected state", exception=None)
        return success

    def map_and_update_modify_lease(self) -> bool:
        """
        Calls the policy to fill a request, with associated state transitions.

        @return boolean success
        """
        success = False
        granted = False
        if self.state == ReservationStates.Failed:
            # Must be a previous failure, or policy marked it as failed. Send update to reset client.
            self.generate_update()
        elif self.state == ReservationStates.Active:
            try:
                self.transition(prefix="modifying lease", state=ReservationStates.Active,
                                pending=ReservationPendingStates.ModifyingLease)
                granted = True
                if granted:
                    success = True
                    self.ticket = self.requested_resources
                    self.logger.debug(f"requested_resources.get_sliver() = {self.requested_resources.get_sliver()}")

                    self.logger.debug(f"approved_resources.get_sliver() = {self.approved_resources.get_sliver()}")

                    if self.requested_resources.get_sliver() is not None:
                        self.approved_resources.set_sliver(sliver=self.requested_resources.get_sliver())

                    self.logger.debug(f"approved_resources.get_sliver() = {self.approved_resources.get_sliver()}")

                    self.resources.update_properties(self, self.approved_resources)
                    self.transition(prefix="modify lease", state=ReservationStates.Active,
                                    pending=ReservationPendingStates.Priming)
            except Exception as e:
                self.logger.error("authority mapper modify e: {}".format(e))
                self.fail_notify(message=str(e))
        else:
            self.fail(message="mapAndUpdateModifyLease: unexpected state")

        return success

    def generate_update(self):
        if self.callback is None:
            self.logger.warning("cannot generate update: no callback")
            return

        try:
            self.update_count += 1
            self.sequence_out += 1
            from fabric_cf.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
            RPCManagerSingleton.get().update_lease(reservation=self)
        except Exception as e:
            self.logger.error("callback failed e:{}".format(e))

    def prepare_extend_lease(self):
        self.requested_resources.validate_incoming_ticket(term=self.requested_term)

    def prepare_modify_lease(self):
        self.requested_resources.validate_incoming_ticket(term=self.requested_term)

    def prepare_probe(self):
        try:
            if self.resources is not None:
                self.resources.prepare_probe()
        except Exception as e:
            self.logger.error("exception in authority prepareProbe e:{}".format(e))

    def probe_pending(self):
        if self.service_pending != ReservationPendingStates.None_:
            self.logger.error("service overrun in probePending")
            return

        self.reap()

        if self.pending_state == ReservationPendingStates.None_:
            if self.is_failed() and not self.notified_about_failure:
                self.generate_update()
                self.notified_about_failure = True

        elif self.pending_state == ReservationPendingStates.Redeeming:
            # We are an authority trying to satisfy a ticket redeem on behalf of a client. Retry policy bind.
            assert self.state == ReservationStates.Ticketed
            if not self.bid_pending and self.map_and_update(extend=False):
                self.logger.debug("Resource assignment (redeem) for #{} completed".format(self.rid))
                self.service_pending = ReservationPendingStates.Redeeming

        elif self.pending_state == ReservationPendingStates.ExtendingLease:
            assert self.state == ReservationStates.Active
            if not self.bid_pending and self.map_and_update(extend=True):
                self.logger.debug("Resource assignment (extend) for #{} completed".format(self.rid))
                self.service_pending = ReservationPendingStates.ExtendingLease

        elif self.pending_state == ReservationPendingStates.ModifyingLease:
            assert self.state == ReservationStates.Active
            if not self.bid_pending and self.map_and_update_modify_lease():
                self.logger.debug("Resource assignment (modify) for #{} completed".format(self.rid))
                self.service_pending = ReservationPendingStates.ModifyingLease

        elif self.pending_state == ReservationPendingStates.Closing:
            if self.resources is None or self.resources.is_closed():
                self.transition(prefix="close complete", state=ReservationStates.Closed,
                                pending=ReservationPendingStates.None_)
                self.pending_recover = False
                self.generate_update()

        elif self.pending_state == ReservationPendingStates.Priming and self.resources.is_active():
            # We are an authority filling a ticket claim. Got resources? Note
            # that active() just means no primes/closes/modifies are still in
            # progress. The primes/closes/modifies could have failed. If
            # something succeeded, then we report what we got as active, else
            # it's a complete bust.
            # If something failed or we are recovering, we need to correct
            # the deficit. For a recovering reservation we need to call
            # correctDeficit regardless of whether there is a real deficit,
            # since the individual nodes may be inconsistent with what the
            # client/broker wanted. For example, they may have the wrong
            # logical ids and resource shares.
            if self.pending_recover or self.get_deficit() != 0:
                # The abstract and the concrete units may be different. We
                # need to adjust the abstract to equal concrete so that
                # future additions of resources will not result in
                # inconsistent abstract unit count.
                self.resources.fix_abstract_units()
                # Policies can instruct us to let go a reservation with a
                # deficit. For example, a policy failed adding resources to
                # the reservation multiple times and it wants to prevent
                # exhausting its inventory from servicing this particular
                # request: probably something is wrong with the request.
                if not self.send_with_deficit:
                    # Call the policy to correct the deficit
                    self.policy.correct_deficit(reservation=self)
                    # XXX: be careful here. we are reusing extending for
                    # the purpose of triggering configuration actions on
                    # the newly assigned nodes. If this is not appropriate,
                    # we may need a new servicePending value
                    self.service_pending = ReservationPendingStates.ExtendingLease
                else:
                    self.pending_recover = False
                    if self.resources.get_units() == 0:
                        message = "no information available"
                        if self.update_data.get_events() is not None:
                            message = self.update_data.get_events()
                        self.fail(message="all units failed priming: {}".format(message))
                    else:
                        self.transition(prefix="prime complete1", state=ReservationStates.Active,
                                        pending=ReservationPendingStates.None_)
                    self.generate_update()
            else:
                self.pending_recover = False
                self.transition(prefix="prime complete2", state=ReservationStates.Active,
                                pending=ReservationPendingStates.None_)
                self.generate_update()

    def service_probe(self):
        # An exception in one of these service routines should mean some
        # unrecoverable, reservation-wide failure. It should not occur, e.g.,
        # if some subset of the resources fail.
        try:
            if self.service_pending == ReservationPendingStates.Redeeming:
                self.service_reserve()

            elif self.service_pending == ReservationPendingStates.ExtendingLease:
                self.service_extend_lease()

            elif self.service_pending == ReservationPendingStates.ModifyingLease:
                self.service_modify_lease()

        except Exception as e:
            self.logger.error("authority failed servicing probe e:{}".format(e))
            self.fail_notify(message="post-op exception: {}".format(e))
        self.service_pending = ReservationPendingStates.None_

    def reap(self):
        """
        Reaps any failed or closed resources. Need more here: if reservation has
        a deficit due to failures then we need to find some replacements.
        @throws Exception
        """
        try:
            if self.resources is not None:
                released = self.resources.collect_released()
                if released is not None:
                    if not released.get_notices().is_empty():
                        self.update_data.post(event=released.get_notices().get_notice())
                    self.policy.release(resources=released)
        except Exception as e:
            self.logger.error("exception in authority reap e: {}".format(e))
            self.logger.error(traceback.format_exc())

    def recover(self):
        if self.state == ReservationStates.Ticketed:
            if self.pending_state == ReservationPendingStates.None_:
                self.actor.redeem(reservation=self, callback=None, caller=None)

            elif self.pending_state == ReservationPendingStates.Redeeming:
                self.transition(prefix="[recover]", state=self.state, pending=ReservationPendingStates.None_)
                self.actor.redeem(reservation=self, callback=None, caller=None)

            elif self.pending_state == ReservationPendingStates.Priming:
                self.pending_recover = True
                self.actor.close(reservation=self)

            elif self.pending_state == ReservationPendingStates.Closing:
                self.actor.close(reservation=self)

            else:
                raise AuthorityException(self.UnexpectedExceptionString.format(self.state, self.pending_state))

        elif self.state == ReservationStates.Active:
            if self.pending_state == ReservationPendingStates.None_:
                self.logger.debug("No op")

            elif self.pending_state == ReservationPendingStates.ExtendingLease:
                self.transition(prefix="[recover]", state=self.state,
                                pending=ReservationPendingStates.None_)
                self.actor.extend_lease(self, None)

            elif self.pending_state == ReservationPendingStates.Priming:
                self.pending_recover = True
                self.actor.close(reservation=self)

            elif self.pending_state == ReservationPendingStates.Closing:
                self.actor.close(reservation=self)

            else:
                raise AuthorityException(self.UnexpectedExceptionString.format(self.state, self.pending_state))
        elif self.state == ReservationStates.Failed:
            self.logger.debug("No op")
        else:
            raise AuthorityException(self.UnexpectedExceptionString.format(self.state, self.pending_state))

    def set_send_with_deficit(self, *, value: bool):
        self.send_with_deficit = value

    def get_deficit(self):
        result = 0
        if self.requested_resources is not None:
            result = self.requested_resources.get_units()

        if self.resources is not None:
            cs = self.resources.get_resources()
            if cs is not None:
                result -= cs.get_units()

        return result

    def get_leased_units(self) -> int:
        if self.resources is not None:
            cs = self.resources.get_resources()
            if cs is not None:
                return cs.get_units()
        return 0

    def get_notices(self) -> str:
        s = super().get_notices()
        if self.resources is not None and self.resources.get_resources() is not None:
            resource_notices = self.resources.get_resources().get_notices()
            if resource_notices is not None:
                s += f" {resource_notices}"

        return s

    def get_ticket(self) -> ResourceSet:
        return self.ticket

    def handle_failed_rpc(self, *, failed: FailedRPC):
        if failed.get_error_type() == RPCError.NetworkError:
            if self.is_failed() or self.is_closed():
                return

        # Resources were allocated successfully but Orchestrator could not be informed
        if failed.get_request_type() == RPCRequestType.UpdateLease and self.resources is not None and \
                self.last_pending_state != ReservationPendingStates.ExtendingLease:
            self.logger.error(f"Closing reservation due to non-recoverable RPC error {failed.get_error_type()}")
            self.actor.close(reservation=self)


class AuthorityReservationFactory:
    """
    Factory class for creating authority reservations
    """
    @staticmethod
    def create(*, resources: ResourceSet, term: Term, slice_obj: ABCSlice, rid: ID, actor: ABCActorMixin = None):
        """
        Create an Authotrity Reservation
        :param resources:
        :param term:
        :param slice_obj:
        :param rid:
        :param actor:
        :return:
        """
        res = AuthorityReservation(rid=rid, resources=resources, term=term, slice_object=slice_obj)
        res.restore(actor=actor, slice_obj=slice_obj)
        return res



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
from datetime import datetime
from enum import Enum
from typing import Dict, Tuple

from fim.graph.abc_property_graph import ABCPropertyGraph
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.capacities_labels import ReservationInfo

from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.apis.abc_kernel_reservation import ABCKernelReservation
from fabric_cf.actor.core.apis.abc_kernel_slice import ABCKernelSlice
from fabric_cf.actor.core.common.exceptions import SliceException
from fabric_cf.actor.core.kernel.slice_state_machine import SliceStateMachine, SliceState, SliceOperation
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.reservation_set import ReservationSet
from fabric_cf.actor.core.util.resource_type import ResourceType
from fabric_cf.actor.fim.fim_helper import FimHelper
from fabric_cf.actor.security.auth_token import AuthToken


class SliceTypes(Enum):
    InventorySlice = 1
    ClientSlice = 2
    BrokerClientSlice = 3


class Slice(ABCKernelSlice):
    """
    Slice implementation. A slice has a globally unique identifier, name,
    description, property list, an owning identity, an access control list, and a
    set of reservations.
    This class is used within the Service Manager, which may hold reservations on
    many sites; on the Broker, which may have provided tickets to the slice for
    reservations at many sites; and on the site Authority, where each slice may
    hold multiple reservations for resources at that site.
    """
    def __init__(self, *, slice_id: ID = None, name: str = "unspecified"):
        # Globally unique identifier.
        self.guid = slice_id
        # Slice name. Not required to be globally or locally unique.
        self.name = name
        # Description string. Has only local meaning.
        self.description = "no description"
        # The slice type: inventory or client.
        self.type = SliceTypes.ClientSlice
        # The owner of the slice.
        self.owner = None
        # Resource type associated with this slice. Used when the slice is used to
        # represent an inventory pool.
        self.resource_type = None
        # The reservations in this slice.
        self.reservations = ReservationSet()
        self.delegations = {}
        # Neo4jGraph Id
        self.graph_id = None
        self.graph = None
        self.state_machine = SliceStateMachine(slice_id=slice_id)
        self.dirty = False
        self.config_properties = None
        self.lock = threading.Lock()
        self.lease_end = None
        self.lease_start = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['reservations']
        del state['delegations']
        del state['graph']
        del state['lock']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.reservations = ReservationSet()
        self.graph = None
        self.delegations = {}
        self.lock = threading.Lock()

    def set_graph_id(self, graph_id: str):
        self.graph_id = graph_id

    def get_graph_id(self) -> str:
        return self.graph_id

    def set_graph(self, *, graph: ABCPropertyGraph):
        self.graph = graph
        self.set_graph_id(graph_id=self.graph.get_graph_id())

    def get_graph(self) -> ABCPropertyGraph:
        return self.graph

    def clone_request(self) -> ABCSlice:
        result = Slice()
        result.name = self.name
        result.guid = self.guid
        return result

    def get_lease_end(self) -> datetime:
        return self.lease_end

    def get_lease_start(self) -> datetime:
        return self.lease_start

    def get_description(self) -> str:
        return self.description

    def get_name(self) -> str:
        return self.name

    def get_owner(self) -> AuthToken:
        return self.owner

    def get_reservations(self) -> ReservationSet:
        return self.reservations

    def get_delegations(self) -> Dict[str, ABCDelegation]:
        return self.delegations

    def get_reservations_list(self) -> list:
        return self.reservations.values()

    def get_resource_type(self) -> ResourceType:
        return self.resource_type

    def get_slice_id(self) -> ID:
        return self.guid

    def is_broker_client(self) -> bool:
        return self.type == SliceTypes.BrokerClientSlice

    def is_client(self) -> bool:
        return not self.is_inventory()

    def is_inventory(self) -> bool:
        return self.type == SliceTypes.InventorySlice

    def is_empty(self) -> bool:
        return self.reservations.is_empty()

    def prepare(self):
        self.reservations.clear()
        self.delegations.clear()
        self.state_machine.clear()
        self.transition_slice(operation=SliceStateMachine.CREATE)

    def register(self, *, reservation: ABCKernelReservation):
        if self.reservations.contains(rid=reservation.get_reservation_id()):
            raise SliceException("Reservation #{} already exists in slice".format(reservation.get_reservation_id()))

        self.reservations.add(reservation=reservation)

    def register_delegation(self, *, delegation: ABCDelegation):
        if delegation.get_delegation_id() in self.delegations:
            raise SliceException("Delegation #{} already exists in slice".format(delegation.get_delegation_id()))

        self.delegations[delegation.get_delegation_id()] = delegation

    def set_lease_end(self, *, lease_end: datetime):
        self.lease_end = lease_end

    def set_lease_start(self, *, lease_start: datetime):
        self.lease_start = lease_start

    def set_broker_client(self):
        self.type = SliceTypes.BrokerClientSlice

    def set_client(self):
        self.type = SliceTypes.ClientSlice

    def set_description(self, *, description: str):
        self.description = description

    def set_inventory(self, *, value: bool):
        if value:
            self.type = SliceTypes.InventorySlice
        else:
            self.type = SliceTypes.ClientSlice

    def get_slice_type(self) -> SliceTypes:
        return self.type

    def set_name(self, *, name: str):
        self.name = name

    def set_owner(self, *, owner: AuthToken):
        self.owner = owner

    def set_resource_type(self, *, resource_type: ResourceType):
        self.resource_type = resource_type

    def soft_lookup(self, *, rid: ID) -> ABCKernelReservation:
        return self.reservations.get(rid=rid)

    def soft_lookup_delegation(self, *, did: str) -> ABCDelegation:
        return self.delegations.get(did, None)

    def __str__(self):
        msg = "{}({})".format(self.name, str(self.guid))
        if self.graph_id is not None:
            msg += " Graph Id:{}".format(self.graph_id)
        if self.owner is not None:
            msg += " Owner:{}".format(self.owner)
        if self.state_machine is not None:
            msg += " State:{}".format(self.state_machine.get_state())
        return msg

    def unregister(self, *, reservation: ABCKernelReservation):
        self.reservations.remove(reservation=reservation)

    def unregister_delegation(self, *, delegation: ABCDelegation):
        if delegation.get_delegation_id() in self.delegations:
            self.delegations.pop(delegation.get_delegation_id())

    def get_state(self) -> SliceState:
        return self.state_machine.get_state()

    def set_dirty(self):
        self.dirty = True

    def clear_dirty(self):
        self.dirty = False

    def is_dirty(self) -> bool:
        return self.dirty

    def transition_slice(self, *, operation: SliceOperation) -> Tuple[bool, SliceState]:
        return self.state_machine.transition_slice(operation=operation, reservations=self.reservations)

    def is_stable_ok(self) -> bool:
        state_changed, slice_state = self.transition_slice(operation=SliceStateMachine.REEVALUATE)

        if slice_state == SliceState.StableOk:
            return True

        return False

    def is_stable_error(self) -> bool:
        state_changed, slice_state = self.transition_slice(operation=SliceStateMachine.REEVALUATE)

        if slice_state == SliceState.StableError:
            return True

        return False

    def is_stable(self) -> bool:
        state_changed, slice_state = self.transition_slice(operation=SliceStateMachine.REEVALUATE)

        if slice_state == SliceState.StableError or slice_state == SliceState.StableOk:
            return True

        return False

    def is_dead_or_closing(self) -> bool:
        state_changed, slice_state = self.transition_slice(operation=SliceStateMachine.REEVALUATE)

        if slice_state == SliceState.Dead or slice_state == SliceState.Closing:
            return True

        return False

    def is_dead(self) -> bool:
        state_changed, slice_state = self.transition_slice(operation=SliceStateMachine.REEVALUATE)

        if slice_state == SliceState.Dead:
            return True

        return False

    def set_config_properties(self, *, value: dict):
        self.config_properties = value

    def get_config_properties(self) -> dict:
        return self.config_properties

    def update_slice_graph(self, sliver: BaseSliver, rid: str, reservation_state: str) -> BaseSliver:
        try:
            self.lock.acquire()
            # Update for Orchestrator for Active / Ticketed Reservations
            if sliver is not None and self.graph_id is not None:
                if sliver.reservation_info is None:
                    sliver.reservation_info = ReservationInfo()
                sliver.reservation_info.reservation_id = rid
                sliver.reservation_info.reservation_state = reservation_state
                FimHelper.update_node(graph_id=self.graph_id, sliver=sliver)
            return sliver
        finally:
            self.lock.release()


class SliceFactory:
    @staticmethod
    def create(*, slice_id: ID, name: str = None, properties: dict = None) -> ABCSlice:
        """
        Create slice
        :param slice_id:
        :param name:
        :param properties:
        :return:
        """
        result = Slice(slice_id=slice_id, name=name)
        result.set_config_properties(value=properties)
        return result